@@ -6,8 +6,6 @@ import os
 
 def EM_fit(model, data, tol = None, maxiter = None, parallel = True):
 
-    if not parallel:
-        os.environ['OMP_NUM_THREADS'] = 1
     if tol is None: 
         tol = 1e-3
     if maxiter is None:
@@ -15,19 +13,10 @@ def EM_fit(model, data, tol = None, maxiter = None, parallel = True):
 
     num_subparts = data["data"].shape[0] #mmm the first dimension of data represents each subpart?? interesting.
     num_resources = len(model["learns"])
-
-    trans_softcounts = np.zeros((num_resources,2,2))
-    emission_softcounts = np.zeros((num_subparts,2,2))
-    init_softcounts = np.zeros((2, 1))
     log_likelihoods = np.zeros((maxiter, 1))
 
-    result = {}
-    result['all_trans_softcounts'] = trans_softcounts
-    result['all_emission_softcounts'] = emission_softcounts
-    result['all_initial_softcounts'] = init_softcounts
-
     for i in range(maxiter):
-        result = E_step.run(data, model, result['all_trans_softcounts'], result['all_emission_softcounts'], result['all_initial_softcounts'], 1)
+        result = E_step.run(data, model, 1, int(parallel))
         for j in range(num_resources):
             result['all_trans_softcounts'][j] = result['all_trans_softcounts'][j].transpose()
         for j in range(num_subparts):
