@@ -1,18 +1,29 @@
+#########################################
+# Model.py                              #
+# Model                                 #
+#                                       #
+# @author Anirudhan Badrinath           #
+# Last edited: 07 April 2021            #
+#########################################
+
 import numpy as np
 import numbers
 import os
 import pandas as pd
 import random
+import pickle
 import urllib.request as urllib2
 from pyBKT.generate import synthetic_data, random_model_uni
 from pyBKT.fit import EM_fit, predict_onestep
 from pyBKT.util import crossvalidate, data_helper, check_data, metrics
 
+pd.options.display.float_format = '{:,.5f}'.format
+
 class Model:
     MODELS_BKT = ['multilearn', 'multiprior', 'multipair', 'multigs']
     MODEL_ARGS = ['parallel', 'num_fits', 'seed', 'defaults'] + MODELS_BKT
     FIT_ARGS = ['skills', 'num_fits', 'defaults',
-                            'parallel', 'forgets'] + MODELS_BKT
+                            'parallel', 'forgets', 'preload'] + MODELS_BKT
     CV_ARGS = FIT_ARGS + ['folds', 'seed']
     DEFAULTS = {'num_fits': 5,
                 'defaults': None,
@@ -29,7 +40,8 @@ class Model:
                     'multilearn': 'template_id',
                     'multiprior': 'correct',
                     'multipair': 'problem_id',
-                    'multigs': 'template_id'}
+                    'multigs': 'template_id',
+                    'folds': 'template_id'}
     INITIALIZABLE_PARAMS = ['prior', 'learns', 'guesses', 'slips', 'forgets']
 
     def __init__(self, **kwargs):
@@ -46,14 +58,15 @@ class Model:
         self.fit_model = None
         self.manual_param_init = False
         self._check_args(Model.MODEL_ARGS, kwargs)
-        self._update_param(['parallel', 'num_fits', 'seed', 'defaults'], kwargs)
-        self._update_param('model_type', self._update_defaults(kwargs))
+        self.keep = {}
+        self._update_param(['parallel', 'num_fits', 'seed', 'defaults'], kwargs, keep = True)
+        self._update_param('model_type', self._update_defaults(kwargs), keep = True)
 
     def fit(self, data_path = None, data = None, **kwargs):
         """
         Fits a BKT model given model and data information. Takes arguments skills,
         number of initialization fits, default column names (i.e. correct, skill_name),
-        parallelization, and model types.
+        parallelization, and model types. Resets model state if uninitialized.
 
         >>> model = Model(seed = 42)
         >>> model.fit(data_path = 'as.csv', forgets = True, skills = 'Box and Whisker')
@@ -61,16 +74,40 @@ class Model:
         0.6128265543747811
 
         """
+        if not self.manual_param_init:
+            self.fit_model = {}
+        self.partial_fit(data_path = data_path, data = data, **kwargs)
+
+    def partial_fit(self, data_path = None, data = None, **kwargs):
+        """
+        Partially fits a BKT model given model and data information. Takes arguments skills,
+        number of initialization fits, default column names (i.e. correct, skill_name),
+        parallelization, and model types. Behaviour is to ignore if the model is changed
+        between partial fits since parameters are copied but data is reprocessed. Note
+        that model type need not be specified when using partial fit after the first partial
+        fit.
+
+        >>> model = Model(seed = 42)
+        >>> model.partial_fit(data_path = 'as.csv', forgets = True, skills = 'Box and Whisker')
+        >>> model.partial_fit(data_path = 'as.csv', forgets = True, skills = 'Box and Whisker')
+        >>> model.partial_fit(data_path = 'as.csv', forgets = True, skills = 'Box and Whisker')
+        >>> model.evaluate(data_path = 'as.csv', metric = 'auc')
+        0.6579800168987382
+
+        """
+        self.manual_param_init = True
         self._check_data(data_path, data)
         self._check_args(Model.FIT_ARGS, kwargs)
         self._update_param(['skills', 'num_fits', 'defaults', 
                             'parallel', 'forgets'], kwargs)
-        self._update_param('model_type', self._update_defaults(kwargs))
-        if not self.manual_param_init:
+        if self.fit_model is None or self.fit_model == {}:
             self.fit_model = {}
+            self._update_param('model_type', self._update_defaults(kwargs))
         all_data = self._data_helper(data_path, data, self.defaults, self.skills, self.model_type)
+        self._update_param(['skills'], {'skills': list(all_data.keys())})
         for skill in all_data:
-            self.fit_model[skill] = self._fit(all_data[skill], skill, self.forgets)
+            self.fit_model[skill] = self._fit(all_data[skill], skill, self.forgets, 
+                                              preload = kwargs['preload'] if 'preload' in kwargs else False)
         self.manual_param_init = False
 
     def predict(self, data_path = None, data = None):
@@ -82,12 +119,23 @@ class Model:
 
         >>> model = Model(seed = 42)
         >>> model.fit(data_path = 'as.csv', forgets = True, skills = 'Box and Whisker')
-        >>> model.predict(data_path = 'as.csv')
-        {'Box and Whisker': (array([0.75036   , 0.60286553, 0.75036   , ..., 0.80416815, 0.83126012,
-               0.77249206]), array([[0.17072973, 0.41991344, 0.17072973, ..., 0.07982386, 0.03405348,
-                0.13333885],
-               [0.82927027, 0.58008656, 0.82927027, ..., 0.92017614, 0.96594652,
-                0.86666115]]))}
+        >>> preds_df = model.predict(data_path = 'as.csv')
+        >>> preds_df[preds_df['skill_name'] == 'Box and Whisker'][['user_id', 'correct', 'correct_predictions', 'state_predictions']]
+              user_id  correct  correct_predictions  state_predictions
+        0       64525        1              0.69205            0.28276
+        1       64525        1              0.80226            0.10060
+        2       70363        0              0.69205            0.28276
+        3       70363        1              0.54989            0.51775
+        4       70363        0              0.74196            0.20028
+        ...       ...      ...                  ...                ...
+        3952    96297        1              0.84413            0.03139
+        3953    96297        1              0.84429            0.03113
+        3954    96297        1              0.84432            0.03108
+        3955    96298        1              0.69205            0.28276
+        3956    96298        1              0.80226            0.10060
+
+        [3957 rows x 4 columns]
+
         """
         self._check_data(data_path, data)
         if self.fit_model is None:
@@ -97,8 +145,9 @@ class Model:
                              model_type = self.model_type, gs_ref = self.fit_model,
                              resource_ref = self.fit_model,
                              return_df = True)
-        df['correct_predictions'] = 0
-        df['state_predictions'] = 0
+        # default best effort prediction of 0.5
+        df['correct_predictions'] = 0.5
+        df['state_predictions'] = 0.5
         for skill in all_data:
             correct_predictions, state_predictions = self._predict(self.fit_model[skill], all_data[skill])
             df.loc[all_data[skill]['index'], 'correct_predictions'] = correct_predictions
@@ -118,15 +167,24 @@ class Model:
 
         """
         self._check_data(data_path, data)
+        if not isinstance(metric, (tuple, list)):
+            metric = [metric]
         if self.fit_model is None:
             raise ValueError("model has not been fitted yet")
-        elif isinstance(metric, str):
-            if not metric in metrics.SUPPORTED_METRICS:
-                raise ValueError("metric must be one of: " + ", ".join(metrics.SUPPORTED_METRICS))
-            metric = getattr(metrics, metric)
+        else:
+            for i in range(len(metric)):
+                m = metric[i]
+                if isinstance(m, str):
+                    if not m in metrics.SUPPORTED_METRICS:
+                        raise ValueError("metric must be one of: " + ", ".join(metrics.SUPPORTED_METRICS))
+                    metric[i] = metrics.SUPPORTED_METRICS[m] 
+                elif not callable(m):
+                    raise ValueError("metric must either be a string, function or list/tuple of strings and functions")
+
         all_data = self._data_helper(data_path, data, self.defaults, self.skills, self.model_type,
                                      gs_ref = self.fit_model, resource_ref = self.fit_model)
-        return self._evaluate(all_data, metric)
+        results = self._evaluate(all_data, metric)
+        return results[0] if len(results) == 1 else results
 
     def crossvalidate(self, data = None, data_path = None, metric = metrics.rmse, **kwargs):
         """
@@ -135,16 +193,42 @@ class Model:
         default column names, parallelization, and model types) as arguments.
 
         >>> model = Model(seed = 42)
-        >>> model.crossvalidate(data_path = 'as.csv', skills = 'Box and Whisker')
-        {'Box and Whisker': [0.42907885779555427, 0.43025621177657264, 0.4009203965037577, 0.4198850275047665, 0.4271669758997527]}
+        >>> model.crossvalidate(data_path = 'as.csv')
+                                                            mean_error
+        skill
+        Circle Graph                                           0.45840
+        Percent Of                                             0.38005
+        Finding Percents                                       0.42757
+        Equivalent Fractions                                   0.45754
+        Proportion                                             0.45437
+        ...                                                        ...
+        Solving Systems of Linear Equations                    0.38976
+        Simplifying Expressions positive exponents             0.46494
+        Parts of a Polyomial, Terms, Coefficient, Monom...     0.33278
+        Finding Slope From Equation                            0.30684
+        Recognize Quadratic Pattern                            0.00000
+
+        [110 rows x 1 columns]
 
         """
-        if data is None and data_path is None:
+        metric_names = []
+        if not isinstance(metric, (tuple, list)):
+            metric = [metric]
+        if not isinstance(data, pd.DataFrame) and not isinstance(data_path, str):
             raise ValueError("no data specified")
-        elif isinstance(metric, str):
-            if not metric in metrics.SUPPORTED_METRICS:
-                raise ValueError("metric must be one of: " + ", ".join(metrics.SUPPORTED_METRICS))
-            metric = getattr(metrics, metric)
+        else:
+            for i in range(len(metric)):
+                m = metric[i]
+                if isinstance(m, str):
+                    if not m in metrics.SUPPORTED_METRICS:
+                        raise ValueError("metric must be one of: " + ", ".join(metrics.SUPPORTED_METRICS))
+                    metric[i] = metrics.SUPPORTED_METRICS[m]
+                    metric_names.append(m)
+                elif callable(m):
+                    metric_names.append(m.__name__)
+                else:
+                    raise ValueError("metric must either be a string, function or list/tuple of strings and functions")
+
         self._check_args(Model.CV_ARGS, kwargs)
         self._update_param(['skills', 'num_fits', 'defaults', 
                             'parallel', 'forgets', 'seed', 'folds'], kwargs)
@@ -152,13 +236,17 @@ class Model:
         metric_vals = {}
         if not self.manual_param_init:
             self.fit_model = {}
-        all_data = self._data_helper(data_path, data, self.defaults, self.skills, self.model_type)
+        if isinstance(self.folds, str):
+            self._update_defaults({'folds': self.folds})
+        all_data = self._data_helper(data_path, data, self.defaults, self.skills, self.model_type, folds = isinstance(self.folds, str))
+        self._update_param(['skills'], {'skills': list(all_data.keys())})
         for skill in all_data:
             metric_vals[skill] = self._crossvalidate(all_data[skill], skill, metric)
         self.manual_param_init = False
-        df = pd.DataFrame(metric_vals)
-        df.columns = ['skill', 'mean_error']
-        return df
+        df = pd.DataFrame(metric_vals.items())
+        df.columns = ['skill', 'dummy']
+        df[metric_names] = pd.DataFrame(df['dummy'].tolist(), index = df.index)
+        return df.set_index('skill').drop(columns = 'dummy')
 
     @property
     def coef_(self):
@@ -209,6 +297,28 @@ class Model:
         """ 
         Returns a DataFrame containing fitted parameters for easy
         printing.
+
+        >>> model = Model(seed = 42)
+        >>> model.fit(data_path = 'as.csv', multilearn = True, forgets = True, skills = 'Box and Whisker')
+        >>> model.params()
+                                          value
+        skill           param   class          
+        Box and Whisker prior   default 0.67443
+                        learns  30799   0.16737
+                                30059   0.33788
+                                30060   0.28723
+                                63448   0.10231
+                                63447   0.07025
+                                63446   0.13453
+                        guesses default 0.31793
+                        slips   default 0.12543
+                        forgets 30799   0.00000
+                                30059   0.04908
+                                30060   0.01721
+                                63448   0.03895
+                                63447   0.00000
+                                63446   0.01058
+
         """
         coefs = self.coef_
         formatted_coefs = []
@@ -216,11 +326,34 @@ class Model:
             for param in coefs[skill]:
                 classes = self._format_param(skill, param, coefs[skill][param])
                 for class_ in classes:
-                   formatted_coefs.append((skill, param, class_, classes[class_])) 
+                   formatted_coefs.append((skill, param, str(class_), classes[class_]))
         df = pd.DataFrame(formatted_coefs)
         df.columns = ['skill', 'param', 'class', 'value']
         return df.set_index(['skill', 'param', 'class'])
 
+    def save(self, loc):
+        """
+        Saves a model to disk. Uses Python pickles.
+
+        >>> model = Model(seed = 42)
+        >>> model.fit(data_path = 'as.csv', multilearn = True, forgets = True, skills = 'Box and Whisker')
+        >>> model.save('model.pkl')
+        """
+        with open(loc, 'wb') as handle:
+            pickle.dump(self, handle, protocol=pickle.HIGHEST_PROTOCOL)
+
+    def load(self, loc):
+        """
+        Loads model given by loc into the current model.
+
+        >>> model = Model(seed = 42)
+        >>> model.load('model.pkl')
+        """
+        with open(loc, 'rb') as handle:
+            orig_model = pickle.load(handle)
+        for attr in vars(orig_model):
+            setattr(self, attr, getattr(orig_model, attr))
+
 
     def fetch_dataset(self, link, loc):
         """
@@ -235,14 +368,14 @@ class Model:
         with open(os.path.normpath(loc + '/' + name), 'wb') as f:
             f.write(file_data.read())
 
-    def _data_helper(self, data_path, data, defaults, skills, model_type, gs_ref = None, resource_ref = None, return_df = False):
+    def _data_helper(self, data_path, data, defaults, skills, model_type, gs_ref = None, resource_ref = None, return_df = False, folds = False):
         """ Processes data given defaults, skills, and the model type. """
-        if data_path is not None:
+        if isinstance(data_path, str):
             data_p = data_helper.convert_data(data_path, skills, defaults = defaults, model_type = model_type, 
-                                              gs_refs = gs_ref, resource_refs = resource_ref, return_df = return_df)
-        elif data is not None:
+                                              gs_refs = gs_ref, resource_refs = resource_ref, return_df = return_df, folds = folds)
+        elif isinstance(data, pd.DataFrame):
             data_p = data_helper.convert_data(data, skills, defaults = defaults, model_type = model_type,
-                                                gs_refs = gs_ref, resource_refs = resource_ref, return_df = return_df)
+                                                gs_refs = gs_ref, resource_refs = resource_ref, return_df = return_df, folds = folds)
         if not return_df:
             for d in data_p.values():
                 check_data.check_data(d)
@@ -251,25 +384,29 @@ class Model:
                 check_data.check_data(d)
         return data_p
 
-    def _fit(self, data, skill, forgets):
+    def _fit(self, data, skill, forgets, preload = False):
         """ Helper function for fitting data. """
         num_learns = len(data["resource_names"])
         num_gs = len(data["gs_names"])
         self._check_manual_param_init(num_learns, num_gs, skill)
         num_fit_initializations = self.num_fits
         best_likelihood = float("-inf")
+        best_model = None
 
         for i in range(num_fit_initializations):
             fitmodel = random_model_uni.random_model_uni(num_learns, num_gs)
-            if self.manual_param_init and skill in self.fit_model:
-                for var in self.fit_model[skill]:
-                    fitmodel[var] = self.fit_model[skill][var]
-
             if forgets:
                 fitmodel["forgets"] = np.random.uniform(size = fitmodel["forgets"].shape)
-            fitmodel, log_likelihoods = EM_fit.EM_fit(fitmodel, data, parallel = self.parallel)
-            if log_likelihoods[-1] > best_likelihood:
-                best_likelihood = log_likelihoods[-1]
+            if self.manual_param_init and skill in self.fit_model:
+                for var in self.fit_model[skill]:
+                    if var in fitmodel:
+                        fitmodel[var] = self.fit_model[skill][var]
+            if not preload:
+                fitmodel, log_likelihoods = EM_fit.EM_fit(fitmodel, data, parallel = self.parallel)
+                if log_likelihoods[-1] > best_likelihood:
+                    best_likelihood = log_likelihoods[-1]
+                    best_model = fitmodel
+            else:
                 best_model = fitmodel
         fit_model = best_model
         fit_model["learns"] = fit_model["As"][:, 1, 0]
@@ -292,31 +429,42 @@ class Model:
             real_data = all_data[skill]['data']
             true = np.append(true, real_data.sum(axis = 0))
             pred = np.append(pred, correct_predictions)
-        return metric(true, pred)
+        true = true - 1
+        try:
+            res = [m(true, pred) for m in metric]
+        except ValueError:
+            res = [m(true, pred.round(0)) for m in metric]
+        return res
 
     def _crossvalidate(self, data, skill, metric):
         """ Helper function for crossvalidating. """
-        return crossvalidate.crossvalidate(self, data, skill, self.folds, metric, self.seed)
+        if isinstance(self.folds, str):
+            return crossvalidate.crossvalidate(self, data, skill, self.folds, metric, self.seed, True)
+        else:
+            return crossvalidate.crossvalidate(self, data, skill, self.folds, metric, self.seed)
 
     def _format_param(self, skill, param, value):
         """ Formats parameter for nice printing. """
         if isinstance(value, np.ndarray):
             ptype = 'resource_names' if (param == 'learns' or param == 'forgets') \
                                      else 'gs_names'
-            return dict(zip(self.fit_model[skill][ptype], value))
+            names = [str(i) for i in self.fit_model[skill][ptype]]
+            return dict(sorted(zip(names, value)))
         else:
             return {'default': value}
 
-    def _update_param(self, params, args):
+    def _update_param(self, params, args, keep = False):
         """ Updates parameters given kwargs. """
         if isinstance(args, dict):
             for param in params:
-                if param not in args:
+                if param not in args and (param not in self.keep or not self.keep[param]):
                     setattr(self, param, Model.DEFAULTS[param])
-                else:
+                elif param in args:
                     setattr(self, param, args[param])
+                self.keep[param] = keep
         else:
             setattr(self, params, args)
+            self.keep[params] = keep
 
     def _update_defaults(self, defaults):
         """ Update the default column names. """
@@ -347,7 +495,7 @@ class Model:
                 valid = valid and isinstance(params[param], numbers.Number)
             else:
                 valid = valid and isinstance(params[param], np.ndarray) \
-                                   and params[param] in Model.INITIALIZABLE_PARAMS
+                                   and param in Model.INITIALIZABLE_PARAMS
         if 'learns' in params and 'forgets' in params:
             valid = valid and (len(params['learns']) == len(params['forgets']))
         if 'guesses' in params and 'slips' in params:
@@ -372,13 +520,11 @@ class Model:
                         ", ".join(expected_args))
 
     def _check_data(self, data_path, data):
-        if not data_path and not data:
+        if not isinstance(data_path, str) and not isinstance(data, pd.DataFrame):
             raise ValueError("no data specified")
-        elif data_path is not None and data is not None:
+        elif isinstance(data_path, str) and isinstance(data, pd.DataFrame):
             raise ValueError("cannot specify both data location and data")
-        elif data is not None and not isinstance(data, pd.DataFrame):
-            raise ValueError("data must be a Pandas DataFrame")
-        elif data_path is not None and not os.path.exists(data_path):
+        elif isinstance(data_path, str) and not os.path.exists(data_path):
             raise ValueError("data path is invalid or file not found")
 
     def __repr__(self):
